@@ -2,7 +2,7 @@
 Name:    py416.files
 Author:  Ezio416
 Created: 2022-08-16
-Updated: 2022-09-16
+Updated: 2022-09-21
 
 Functions for file system manipulation
 OS-agnostic (Windows/Unix) - Windows paths will always have forward slashes
@@ -30,20 +30,20 @@ class File():
         return os.path.getatime(self.path)
 
     @property
-    def atimes(self) -> list:
-        return dt.fromtimestamp(self.atime).strftime('%Y,%m,%d,%H,%M,%S,%f').split(',')
+    def atimes(self) -> tuple:
+        return tuple(dt.fromtimestamp(self.atime).strftime('%Y,%m,%d,%H,%M,%S,%f').split(','))
 
     @property
-    def children(self) -> list:
-        return listdir(self.path) if self.isdir else []
+    def children(self) -> tuple:
+        return listdir(self.path) if self.isdir else ()
 
     @property
     def ctime(self) -> float:
         return os.path.getctime(self.path)
 
     @property
-    def ctimes(self) -> list:
-        return dt.fromtimestamp(self.ctime).strftime('%Y,%m,%d,%H,%M,%S,%f').split(',')
+    def ctimes(self) -> tuple:
+        return tuple(dt.fromtimestamp(self.ctime).strftime('%Y,%m,%d,%H,%M,%S,%f').split(','))
 
     @property
     def exists(self) -> bool:
@@ -66,8 +66,8 @@ class File():
         return os.path.getmtime(self.path)
 
     @property
-    def mtimes(self) -> list:
-        return dt.fromtimestamp(self.mtime).strftime('%Y,%m,%d,%H,%M,%S,%f').split(',')
+    def mtimes(self) -> tuple:
+        return tuple(dt.fromtimestamp(self.mtime).strftime('%Y,%m,%d,%H,%M,%S,%f').split(','))
 
     @property
     def name(self) -> str:
@@ -78,7 +78,7 @@ class File():
         return File(parent(self.path))
 
     @property
-    def parts(self) -> list:
+    def parts(self) -> tuple:
         return splitpath(self.path)
 
     @property
@@ -141,13 +141,28 @@ def cd(path:str='..') -> str:
         - Default: up a directory
     - Return: `str` with current working directory (formatted with `/`)
     '''
-    if gettype(path) != 'str':
-        raise TypeError('input must be a string')
     path = getpath(path)
     makedirs(path)
     os.chdir(path)
     return path
 
+def checkwindrive(drive:str) -> str:
+    '''
+    - Checks if a given string is a Windows drive letter (i.e. 'C:', 'D:\\\\', 'E:/')
+    - Input: `drive` (`str`): string to check
+    - Return: `str` with drive (formatted with `/`)
+    '''
+    drive = forslash(drive).upper()
+    if len(drive) not in (2, 3):
+        return ''
+    if drive[0] not in 'ABCDEFGHIJKLMNOPQRSTUVWXYZ' or drive[1] != ':':
+        return ''
+    if len(drive) == 3:
+        if drive[2] != '/':
+            return ''
+        return drive
+    return drive + '/'
+
 def checkzip(path:str) -> bool:
     '''
     - Checks if an archive file (.7z or .zip) exists and is valid
@@ -158,8 +173,7 @@ def checkzip(path:str) -> bool:
         - `True`: file exists and is valid
         - `False`: file doesn't exist, possibly because we deleted it due to corruption
     '''
-    if gettype(path) != 'str':
-        raise TypeError('input must be a string')
+    path = getpath(path)
     try:
         if path.endswith('.7z'):
             from py7zr import exceptions, SevenZipFile
@@ -188,15 +202,15 @@ def copymove(func):
     - Copy and move are very similar, so this handles some of what they share
     '''
     def _copymove(path:str, dest:str, overwrite:bool=False):
-        if gettype(path) != gettype(dest) != 'str':
-            raise TypeError('input must be a string')
-        if not os.path.exists(path):
-            raise FileNotFoundError('path does not exist')
-        if os.path.exists(dest) and not os.path.isdir(dest):
-            raise FileExistsError('you can\'t move something into a file')
-        if getpath(path) == getpath(dest):
-            raise ValueError('you can\'t move something into itself')
+        path = getpath(path)
+        dest = getpath(dest)
         overwrite = bool(overwrite)
+        if not os.path.exists(path):
+            raise FileNotFoundError(f'not found: {path}')
+        if path == dest:
+            raise ValueError(f'you can\'t move something into itself: {path}')
+        if os.path.exists(dest) and not os.path.isdir(dest):
+            raise FileExistsError(f'you can\'t move something into a file: {dest}')
         makedirs(dest)
         return forslash(func(path, dest, overwrite))
     return _copymove
@@ -214,17 +228,17 @@ def copy(path:str, dest:str, overwrite:bool=False) -> str:
             - Default: `False`
     - Return: `str` with path to copied file/directory (formatted with `/`)
     '''
-    newpath = f'{dest}/{os.path.basename(path)}'
-    newpath_exists = os.path.exists(newpath)
+    new_path = f'{dest}/{os.path.basename(path)}'
+    new_path_exists = os.path.exists(new_path)
     if os.path.isfile(path): # copying file
-        if newpath_exists:
-            if overwrite:
-                return sh.copy2(path, newpath) # overwriting dest file
-            raise FileExistsError('destination file already exists') # not overwriting
-        return sh.copy2(path, newpath) # dest file doesn't exist, good
-    if newpath_exists and overwrite: # copying directory
-        return sh.copytree(path, newpath, dirs_exist_ok=True) # overwriting dest dir
-    return sh.copytree(path, newpath) # dest dir doesn't exist, or if it does, shutil raises FileExistsError
+        if new_path_exists and not overwrite:
+            raise FileExistsError(f'destination file already exists: {new_path}')
+        return sh.copy2(path, new_path)
+    if new_path_exists: # copying directory
+        if overwrite:
+            return sh.copytree(path, new_path, dirs_exist_ok=True) # overwriting dest dir
+        raise FileExistsError(f'destination directory already exists: {new_path}')
+    return sh.copytree(path, new_path) # dest dir doesn't exist, good
 
 def delete(path:str, force:bool=False) -> None:
     '''
@@ -233,11 +247,10 @@ def delete(path:str, force:bool=False) -> None:
         - `path` (`str`): path to file or directory
         - `force` (`bool`): whether to try `shutil.rmtree()` to delete a directory
     '''
-    if gettype(path) != 'str':
-        raise TypeError('input must be a string')
-    if not os.path.exists(path):
-        raise FileNotFoundError('path does not exist')
+    path = getpath(path)
     force = bool(force)
+    if not os.path.exists(path):
+        raise FileNotFoundError(f'not found: {path}')
     if os.path.isdir(path):
         if force:
             sh.rmtree(path)
@@ -254,7 +267,7 @@ def forslash(path:str) -> str:
     - Return: `str` with path (formatted with `/`)
     '''
     if gettype(path) != 'str':
-        raise TypeError('input must be a string')
+        raise TypeError(f'input must be a string; invalid: {path}')
     return path.replace('\\', '/')
 
 def getcwd() -> str:
@@ -272,14 +285,47 @@ def getpath(path:str) -> str:
         - If relative, we assume it's in the current working directory
     - Return: `str` with path (formatted with `/`)
     '''
-    if gettype(path) != 'str':
-        raise TypeError('input must be a string')
-    drives = ['/'] + [f'{ch}:/' for ch in 'abcdefghijklmnopqrstuvwxyz']
-    if splitpath(path)[0].lower() not in drives: # relative path
-        return f'{getcwd()}/{forslash(path)}'
-    return forslash(os.path.abspath(path))
+    parts = list(splitpath(path))
+    if parts == ['']: # special case
+        return ''
+    path = joinpath(parts)
+    root = parts[0]
+    if root.startswith('//') or root == '/': # Windows network location or Unix root
+        return path
+    cwdrive = checkwindrive(root)
+    if cwdrive: # Windows root
+        parts[0] = cwdrive
+        return joinpath(parts)
+    return f'{getcwd()}/{path}'
 
-def listdir(path:str='', dirs:bool=True, files:bool=True) -> list:
+def joinpath(*parts) -> str:
+    '''
+    - Imitator for `os.path.join()`
+    - Joins parts of a path together in the order they were received
+    - Input: `parts` (iterable): directories/file to join together
+    - Return: `str` with path (formatted with `/`)
+    '''
+    parts = list(unpack([list(splitpath(part)) for part in unpack(parts)])) # split elements if partial paths
+    if parts == ['/', '/']: # special case
+        return '/'
+    while '' in parts:
+        parts.remove('') # special case, such as passing 'folder/..' to splitpath()
+    if not len(parts):
+        return '' # nothing was passed, or it was cancelled out with '..'
+    if parts[0].startswith('//'):
+        return '/'.join(parts) # Windows network location
+    if parts == ['/']:
+        return '/' # Unix root, alone
+    if parts[0] == '/':
+        return '/' + '/'.join(parts[1:]) # Unix root, preceding
+    cwdrive = checkwindrive(parts[0])
+    if cwdrive:
+        if len(parts) == 1:
+            return cwdrive # Windows drive root, alone
+        return cwdrive + '/'.join(parts[1:]) # Windows drive root, preceding
+    return '/'.join(parts)
+
+def listdir(path:str='.', dirs:bool=True, files:bool=True) -> tuple:
     '''
     - Wrapper for `os.listdir()`
     - Lists directories/files within a directory
@@ -290,25 +336,23 @@ def listdir(path:str='', dirs:bool=True, files:bool=True) -> list:
             - Default: `True`
         - `files` (`bool`): whether to list files
             - Default: `True`
-    - Return: `list` of `str` with paths (formatted with `/`)
+    - Return: `tuple` of `str` with paths (formatted with `/`)
     '''
-    if gettype(path) != 'str':
-        raise TypeError('input must be a string')
-    if not os.path.exists(path):
-        return []
+    path = getpath(path)
     dirs = bool(dirs)
     files = bool(files)
+    if not os.path.exists(path):
+        raise FileNotFoundError(f'not found: {path}')
     result = []
-    path = forslash(path) if path else getcwd()
     for child in os.listdir(path):
-        child = f'{path}/{child}'
+        child = joinpath(path, child)
         if dirs and os.path.isdir(child):
             result.append(child)
         if files and not os.path.isdir(child):
             result.append(child)
-    return result
+    return tuple(result)
 
-def log(path:str, msg:str, ts:bool=True, ts_args:list=[1,0,1,1,1,0]) -> None:
+def log(path:str, msg:str, ts:bool=True, ts_args:tuple=(1,0,1,1,1,0)) -> None:
     '''
     - Logs to file with current timestamp
     - Creates file and its parent directory if nonexistent
@@ -320,11 +364,12 @@ def log(path:str, msg:str, ts:bool=True, ts_args:list=[1,0,1,1,1,0]) -> None:
         - `ts_args` (`list`/`tuple`): arguments to pass to `py416.timestamp()`
             - Default example: [2022-08-19 13:24:54 -06:00]
     '''
-    if gettype(path) != gettype(msg) != 'str':
-        raise ValueError('input must be a string')
-    if gettype(ts_args) not in ['list', 'tuple']:
-        raise ValueError('input must be a list/tuple')
+    path = getpath(path)
+    if gettype(msg) != 'str':
+        raise ValueError(f'input must be a string; invalid: {msg}')
     ts = bool(ts)
+    if gettype(ts_args) not in ('list', 'tuple'):
+        raise ValueError(f'input must be a list/tuple; invalid: {ts_args}')
     makedirs(parent(path))
     now = timestamp(*ts_args) + '  ' if ts else ''
     orig_stdout = sys.stdout
@@ -337,7 +382,7 @@ def log(path:str, msg:str, ts:bool=True, ts_args:list=[1,0,1,1,1,0]) -> None:
     finally:
         sys.stdout = orig_stdout
 
-def makedirs(*dirs, ignore_errors:bool=True) -> list:
+def makedirs(*dirs, ignore_errors:bool=True) -> tuple:
     '''
     - Wrapper for `os.makedirs()`
     - Creates directories if nonexistent
@@ -348,16 +393,17 @@ def makedirs(*dirs, ignore_errors:bool=True) -> list:
         - `ignore_errors` (`bool`): whether to catch all Exceptions in the process
             - Useful to create as many of the requested directories as possible
             - Default: `True`
-    - Return: `list` of directories we failed to create
+    - Return: `tuple` of directories we failed to create
     '''
-    if gettype(dirs) not in ['list', 'str', 'tuple']:
-        raise TypeError('input must be a string, list, or tuple')
+    if gettype(dirs) not in ('list', 'str', 'tuple'):
+        raise TypeError(f'input must be a string/list/tuple; invalid: {dirs}')
     ignore_errors = bool(ignore_errors)
     errored = []
     for dir in unpack(dirs):
         if gettype(dir) != 'str':
             errored.append(dir)
             continue
+        dir = getpath(dir)
         if not os.path.exists(dir):
             try:
                 os.makedirs(dir)
@@ -365,7 +411,7 @@ def makedirs(*dirs, ignore_errors:bool=True) -> list:
                 if not ignore_errors:
                     raise
                 errored.append(dir)
-    return errored
+    return tuple(errored)
 
 @copymove
 def move(path:str, dest:str, overwrite:bool=False) -> str:
@@ -380,22 +426,19 @@ def move(path:str, dest:str, overwrite:bool=False) -> str:
             - Default: `False`
     - Return: `str` with path to moved file/directory (formatted with `/`)
     '''
-    newpath = f'{dest}/{os.path.basename(path)}'
-    newpath_exists = os.path.exists(newpath)
-    _move = lambda: sh.move(path, dest)
+    new_path = f'{dest}/{os.path.basename(path)}'
+    new_path_exists = os.path.exists(new_path)
     if os.path.isfile(path): # moving file
-        if newpath_exists:
-            if overwrite:
-                return _move() # overwriting dest file
-            raise FileExistsError('destination file already exists') # not overwriting
-        return _move() # dest file doesn't exist, good
-    if newpath_exists: # moving directory
+        if new_path_exists and not overwrite:
+            raise FileExistsError(f'destination file already exists: {new_path}')
+        return sh.move(path, dest)
+    if new_path_exists: # moving directory
         if overwrite:
-            result = sh.copytree(path, newpath, dirs_exist_ok=True) # overwriting dest dir
+            result = sh.copytree(path, new_path, dirs_exist_ok=True) # overwriting dest dir
             delete(path, force=True) # deleting original (we're doing copy->delete manually)
             return result
-        raise FileExistsError('destination directory already exists') # not overwriting
-    return _move() # dest dir doesn't exist, good
+        raise FileExistsError(f'destination directory already exists: {new_path}')
+    return sh.move(path, dest) # dest dir doesn't exist, good
 
 def parent(path:str) -> str:
     '''
@@ -403,25 +446,14 @@ def parent(path:str) -> str:
     - Input: `path` (`str`): path to find the parent of
     - Return: `str` with path (formatted with `/`)
     '''
-    if gettype(path) != 'str':
-        raise TypeError('input must be a string')
-    dirname = lambda _path: getpath(os.path.abspath(f'{getpath(_path)}/..'))
-    if getattr(sys, 'frozen', False):
-        return dirname(sys.executable)
-    try:
-        return dirname(path)
-    except NameError:
-        return getcwd()
-
-def pathjoin(*parts) -> str:
-    '''
-    - Imitator for `os.path.join()`
-    - Joins parts of a path into a string path
-    - Input: `parts` (iterable): directories/file to join together
-    - Return: `str` with path (formatted with `/`)
-    '''
-    parts = [str(item).replace('/', '').replace('\\', '') for item in unpack(parts)]
-    return '/' if parts == [''] else '/'.join(parts)
+    path = getpath(path)
+    if not path:
+        return ''
+    if any([path == '/', # Root
+            checkwindrive(path),
+            path.startswith('//') and len(splitpath(path)) == 1]):
+        return path
+    return getpath(f'{path}/..')
 
 def rename(path:str, name:str) -> str:
     '''
@@ -432,10 +464,11 @@ def rename(path:str, name:str) -> str:
         - `name` (`str`): new basename for file (not path)
     - Return: `str` with new path (formatted with `/`)
     '''
-    if gettype(path) != gettype(name) != 'str':
-        raise TypeError('input must be a string')
+    path = getpath(path)
+    if gettype(name) != 'str':
+        raise TypeError(f'input must be a string; invalid: {name}')
     if not os.path.exists(path):
-        raise FileNotFoundError('path does not exist')
+        raise FileNotFoundError(f'not found: {path}')
     newpath = f'{parent(path)}/{name}'
     os.rename(path, newpath)
     return newpath
@@ -450,39 +483,66 @@ def rmdir(path:str, delroot:bool=True) -> int:
             - Default: `True`
     - Return: `int` with number of deleted directories
     '''
-    if gettype(path) != 'str':
-        raise TypeError('input must be a string')
-    if not os.path.exists(path):
-        raise FileNotFoundError('path does not exist')
-    if not os.path.isdir(path):
-        raise ValueError('path is not a directory')
+    path = getpath(path)
     delroot = bool(delroot)
+    if not os.path.exists(path):
+        raise FileNotFoundError(f'not found: {path}')
+    if not os.path.isdir(path):
+        raise ValueError(f'not a directory: {path}')
     count = 0
     for item in listdir(path):
         if os.path.isdir(item):
             count += rmdir(item)
     if not len(listdir(path)) and delroot:
-        if getpath(path) == getcwd():
+        if path == getcwd():
             cd() # we're in the directory we're trying to delete, so go up
         os.rmdir(path)
         count += 1
     return count
 
-def splitpath(path:str) -> list:
+def splitpath(path:str) -> tuple:
     '''
     - Splits a path string into its parts
     - Input: `path` (`str`): path
-    - Return: `list` of directories/file
+    - Return: `tuple` of directories/file
     '''
-    if gettype(path) != 'str':
-        raise TypeError('input must be a string')
+    if not path:
+        return '',
     path = forslash(path)
-    parts = forslash(os.path.abspath(path)).split('/')
-    parts[0] = f'{parts[0]}/' # root
+    win_net = False
+    
+    if path in ('/', '/.', '/..'): # special cases
+        return '/',
+    parts = path.split('/')
+    if path == '.': # current directory
+        path = getcwd()
+    elif path == '..': # parent of current directory
+        path = forslash(os.path.dirname(getcwd()))
+    elif parts[-1] == '.': # path/. is just path
+        path = path[:-2]
+    elif parts[-1] == '..': # parent of path
+        if len(parts) == 2: # 'folder/..'
+            return '',
+        if path.startswith('//'): # Windows network location
+            path = path.lstrip('/')
+            win_net = True
+        path = os.path.dirname(os.path.dirname(path))
+        if win_net:
+            path = '//' + path
+    parts = path.split('/')
+
+    if not parts[0]: # Unix root
+        parts[0] = '/'
+    cwdrive = checkwindrive(parts[0])
+    if cwdrive: # Windows drive root
+        parts[0] = cwdrive
     if path.startswith('//'): # Windows network location
-        result = [f'//{parts[2]}'] # network root
-        return result + [a for a in parts[3:]] if len(parts) > 2 else result
-    return [parts[0]] if parts[1] == '' else parts
+        result = [f'//{parts[2]}']
+        return tuple(result + [a for a in parts[3:]]) if len(parts) > 2 else tuple(result)
+    if len(parts) == 2:
+        if parts[1] == '':
+            return parts[0],
+    return tuple(parts)
 
 def unzip(path:str, remove:bool=False) -> None:
     '''
@@ -492,11 +552,10 @@ def unzip(path:str, remove:bool=False) -> None:
         - `remove` (`bool`): whether to delete archive after unzipping
             - Default: `False`
     '''
-    if gettype(path) != 'str':
-        raise TypeError('input must be a string')
-    if not os.path.exists(path):
-        raise FileNotFoundError('path does not exist')
+    path = getpath(path)
     remove = bool(remove)
+    if not os.path.exists(path):
+        raise FileNotFoundError(f'not found: {path}')
     fparent = parent(path)
     if path.endswith('.7z'):
         from py7zr import unpack_7zarchive
@@ -508,26 +567,24 @@ def unzip(path:str, remove:bool=False) -> None:
         if remove:
             delete(path)
     else:
-        raise NotImplementedError('unsupported archive format')
+        raise NotImplementedError(f'unsupported archive format: {path.split(".")[-1]}')
 
-def unzipdir(path:str='.', ignore_errors:bool=True) -> int:
+def unzipdir(path:str, ignore_errors:bool=True) -> int:
     '''
-    - Unzips all archives in a directory until it is unable to continue
+    - Unzips all archives in a directory (only 1st level) until it is unable to continue
     - Deletes all archives as it unzips
     - Supports archives of type: (.7z, .gz, .rar, .tar, .zip)
     - Input:
-        - `path` (`str`): directory from which to unzip archives
-            - Default: current working directory
+        - `path` (`str`): directory containing archive files
         - `ignore_errors` (`bool`): whether to catch all Exceptions in unzipping
             - Useful to unzip everything possible in the directory
             - Default: `True`
     - Return: `int` with number of unzipped archives
     '''
-    if gettype(path) != 'str':
-        raise TypeError('input must be a string')
-    if not os.path.exists(path):
-        raise FileNotFoundError('path does not exist')
+    path = getpath(path)
     ignore_errors = bool(ignore_errors)
+    if not os.path.exists(path):
+        raise FileNotFoundError(f'not found: {path}')
     unzipped = 0
     while True:
         unzipped_this_run = 0
